"""Concrete classes for diagnostic objects, for use in standard Python toolchain.
"""

from ._base import Diagnostic, DiagnosticStyle


class DiagnosticError(Diagnostic, Exception):
    """An exception that requires additional diagnostic information to be provided.

    This is a subclass of Exception and can be used as a normal exception, with
    the typical exception handling mechanisms. It can also be printed using
    `rich` to get a pretty presentation of the error.
    """

    style = DiagnosticStyle(
        name="error",
        color="red",
        ascii_symbol="x",
        unicode_symbol="×",
    )


class DiagnosticWarning(Diagnostic, Warning):
    """A warning that requires additional diagnostic information to be provided.

    This is a subclass of Warning and can be used as a normal warning, with
    the typical"""

    style = DiagnosticStyle(
        name="warning",
        color="yellow",
        ascii_symbol="!",
        unicode_symbol="!",
    )


# Don't expose the private module name.
DiagnosticError.__module__ = "diagnostic"
DiagnosticWarning.__module__ = "diagnostic"



"""Present errors that contain causes better understand what happened.
"""

from ._base import Diagnostic, DiagnosticStyle
from ._concrete import DiagnosticError, DiagnosticWarning

__all__ = ["DiagnosticStyle", "Diagnostic", "DiagnosticError", "DiagnosticWarning"]
__version__ = "1.0.0"



"""The main functional unit of this package.
"""

from __future__ import annotations

import dataclasses
import re
import textwrap
from typing import ClassVar, Iterator, Sequence

import rich
import rich.console
import rich.text

RE_code = re.compile(
    r"""
    ^                         # start
        [a-zA-Z][a-zA-Z0-9]*  # with "name" segment
        (
            -                 # then a dash
            [a-zA-Z0-9]+      # followed by "name-or-number" segment
        )*                    # allow multiple segments, separated by dashes
    $                         # end
    """,
    re.VERBOSE,
)
_RE_TAGS = re.compile(
    r"""((\\*)\[([a-z#/@][^[]*?)])""",  # copied over from rich
    re.VERBOSE,
)


def _is_valid_code(s: str) -> bool:
    return re.match(RE_code, s) is not None


def _to_plain_str(s: str | rich.text.Text) -> str:
    """Convert user inputs to a string."""
    if isinstance(s, str):
        return rich.text.Text.from_markup(s).plain
    else:
        return s.plain


def _index_prefix_rich(
    s: str | rich.text.Text,
    console: rich.console.Console,
    *,
    prefix: str,
    indent: str,
) -> rich.text.Text:
    if isinstance(s, str):
        text = console.render_str(s)
    else:
        text = s

    lines = text.split(allow_blank=True)
    body = console.render_str(f"\n{indent}", overflow="ignore").join(lines)
    return console.render_str(prefix, overflow="ignore") + body


def _indent_prefix(s: str | rich.text.Text, *, prefix: str, indent: str) -> str:
    first, _, rest = _to_plain_str(s).partition("\n")
    return "\n".join(filter(None, [prefix + first, textwrap.indent(rest, indent)]))


@dataclasses.dataclass(frozen=True)
class DiagnosticStyle:
    name: str
    color: str
    ascii_symbol: str
    unicode_symbol: str


class Diagnostic:
    """An object that holds diagnostic information to present to a reader."""

    style: ClassVar[DiagnosticStyle]
    """Data about how this diagnostic should be presented"""

    docs_index: ClassVar[str | None] = None
    """
    URL to the documentation index page(s). Must contain a "{code}" placeholder,
    which will be replaced with the code of this instance.
    """

    code: str | None = None
    """
    A unique code to help readers identify this in output, documentation, etc.

    This should be a string of the form "name[-name-or-number]*", where "name" is
    a string of letters and numbers, and "name-or-number" is a string of letters,
    numbers, and dashes.
    """

    message: str | rich.text.Text
    """A short description."""

    causes: Sequence[str | rich.text.Text]
    """A list of strings describing the causes."""

    hint_stmt: str | rich.text.Text | None
    """A hint for what the reader might want to do next."""

    note_stmt: str | rich.text.Text | None
    """
    A note with additional information, that's not part of the "causes" but
    could be useful to know for the reader.
    """

    details_link: str | None
    """
    A link to more details about the problem.

    This is determined automatically if :attr:`code` is set, and
    :attr:`docs_index` is set.
    """

    def __init__(
        self,
        *,
        code: str | None = None,
        message: str | rich.text.Text,
        causes: list[str] | list[rich.text.Text] | list[str | rich.text.Text],
        hint_stmt: str | rich.text.Text | None,
        note_stmt: str | rich.text.Text | None = None,
    ) -> None:
        """
        :param code: Maps to :attr:`code`.
        :param message: Maps to :attr:`message`.
        :param causes: Maps to :attr:`causes`.
        :param hint_stmt: Maps to :attr:`hint_stmt`.
        :param note_stmt: Maps to :attr:`note_stmt`.
        """
        super().__init__()

        if code is None:
            if self.__class__.code is None:
                raise ValueError(
                    f"Cannot create {self.__class__.__name__} object: "
                    "`code` must be provided!"
                )
            code = self.__class__.code
        if not _is_valid_code(code):
            raise ValueError(
                f"Cannot create {self.__class__.__name__} object: "
                f"error code {code!r} must be kebab-case and start "
                "with a character!"
            )
        if not isinstance(causes, list):  # type: ignore
            raise TypeError(
                f"Cannot create {self.__class__.__name__} object: "
                f"`causes` must be a list, not {type(causes).__name__}!"
            )
        if not hasattr(self.__class__, "style"):
            raise TypeError(
                f"Cannot create {self.__class__.__name__} object: "
                "`style` must be set in a subclass of Diagnostic!"
            )

        self.code = code

        self.message = message
        self.causes = causes

        self.note_stmt = note_stmt
        self.hint_stmt = hint_stmt

        if self.__class__.docs_index is not None:
            if "{code}" not in self.__class__.docs_index:
                raise ValueError(
                    f"Cannot create {self.__class__.__name__} object: "
                    "`docs_index` must contain a {code} placeholder!"
                )
            self.details_link: str | None = self.__class__.docs_index.format(
                code=self.code
            )
        else:
            self.details_link = None

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__}("
            f"code={self.code!r}, "
            f"message={self.message!r}, "
            f"causes={self.causes!r}, "
            f"note_stmt={self.note_stmt!r}, "
            f"hint_stmt={self.hint_stmt!r}, "
            f"details_link={self.details_link!r}"
            ")>"
        )

    def __str__(self) -> str:
        return "\n".join(self._str_parts())

    def _str_parts(self) -> Iterator[str]:
        assert self.code is not None
        yield self.code
        yield ""
        yield _to_plain_str(self.message)
        if self.causes:
            yield ""
            yield "Caused by:"
            for item in self.causes:
                yield _indent_prefix(item, prefix="--> ", indent="    ")
        if self.note_stmt is not None or self.hint_stmt is not None:
            yield ""
        if self.note_stmt is not None:
            yield _indent_prefix(self.note_stmt, prefix="note: ", indent="      ")
        if self.hint_stmt is not None:
            yield _indent_prefix(self.hint_stmt, prefix="hint: ", indent="      ")
        if self.details_link is not None:
            yield ""
            yield f"For more details, see {self.details_link}"

    def __rich_console__(
        self,
        console: rich.console.Console,
        options: rich.console.ConsoleOptions,
    ) -> rich.console.RenderResult:
        yield (
            f"[{self.style.color} bold]{self.style.name}[/]: " f"[bold]{self.code}[/]"
        )
        yield ""

        if not options.ascii_only:
            # Present the main message, with relevant causes indented.
            if self.causes:
                yield _index_prefix_rich(
                    self.message,
                    console,
                    prefix=f"[{self.style.color}]{self.style.unicode_symbol}[/] ",
                    indent=f"[{self.style.color}]│[/] ",
                )
                for item in self.causes[:-1]:
                    yield _index_prefix_rich(
                        item,
                        console,
                        prefix=f"[{self.style.color}]├─>[/] ",
                        indent=f"[{self.style.color}]│  [/] ",
                    )
                yield _index_prefix_rich(
                    self.causes[-1],
                    console,
                    prefix=f"[{self.style.color}]╰─>[/] ",
                    indent=f"[{self.style.color}]   [/] ",
                )
            else:
                yield _index_prefix_rich(
                    self.message,
                    console,
                    prefix=f"[{self.style.color}]×[/] ",
                    indent="  ",
                )
        else:
            yield self.message
            if self.causes:
                yield ""
                for item in self.causes:
                    yield item

        if not (self.note_stmt is None and self.hint_stmt is None):
            yield ""

        if self.note_stmt is not None:
            yield _index_prefix_rich(
                self.note_stmt,
                console,
                prefix="[magenta bold]note[/]: ",
                indent="      ",
            )
        if self.hint_stmt is not None:
            yield _index_prefix_rich(
                self.hint_stmt,
                console,
                prefix="[cyan bold]hint[/]: ",
                indent="      ",
            )

        if self.details_link is not None:
            yield ""
            yield f"For more details, see {self.details_link}"



"""Check that all the errors in the source code have a documentation entry.
"""

from __future__ import annotations

import argparse
import sys
from pathlib import Path
from typing import Any, Callable, Iterable, Sequence, TypeVar

import rich
import rich.traceback
from rich.markup import escape

from . import DiagnosticError
from ._parsers import find_code_headings_in_document, find_codes_in_sources

rich.traceback.install(show_locals=True)
T = TypeVar("T")


def _format_to_lines(
    names: set[str],
    codes: dict[str, list[tuple[Path, int]]] | dict[str, list[int]],
    *,
    kind: str,
    fallback_filename: str = "<unset>",
) -> str:
    """Format the codes to lines of text."""
    if not names:
        return ""

    lines = [f"[red]{len(names)} {kind}[/]:"]
    for name in sorted(names):
        lines.append(f"  [magenta]{escape(name)}[/]")
        for entry in codes[name]:
            if isinstance(entry, tuple):
                file, lineno = entry
                filename = str(file)
            else:
                lineno = entry
                filename = fallback_filename
            lines.append(f"    from [blue]{escape(filename)}[/]:[cyan]{lineno}[/]")
    return "\n".join(lines)


def _process(
    source: Path, docs_index: Path, verbose: bool, fail_on_extra: bool
) -> None:
    """Main entry point for the script."""
    code_codes = find_codes_in_sources(source)
    doc_codes = find_code_headings_in_document(docs_index)

    rich.print(f"Found {len(code_codes)} codes in the source code.")
    rich.print(f"Found {len(doc_codes)} codes in the documentation.")
    if verbose:
        rich.get_console().rule()
        rich.print("codes in the source code")
        for code, locations in code_codes.items():
            rich.print(f"  [green]{escape(code)}[/]")
            for filename, lineno in locations:
                rich.print(f"    [blue]{escape(str(filename))}[/]:[cyan]{lineno}[/]")

        rich.get_console().rule()
        rich.print(f"Headings in the {escape(str(docs_index))}")
        for code, linenos in doc_codes.items():
            rich.print(f"  [green]{escape(code)}[/]: {escape(repr(linenos))}")
        rich.get_console().rule()

    undocumented_codes = set(code_codes) - set(doc_codes)
    if not fail_on_extra:
        extra_codes: set[str] = set()
    else:
        extra_codes = set(doc_codes) - set(code_codes)

    if not undocumented_codes:
        rich.print("[bold green]All error codes in code are documented![/] :tada:")
    if not extra_codes and fail_on_extra:
        rich.print(
            "[bold green]All error codes in documentation exist in code![/] :tada:"
        )
    if not undocumented_codes and not extra_codes:
        return

    sections = [
        _format_to_lines(
            undocumented_codes,
            code_codes,
            kind="undocumented",
        ),
        _format_to_lines(
            extra_codes,
            doc_codes,
            kind="extra",
            fallback_filename=str(docs_index),
        ),
    ]
    causes = list(filter(None, sections))

    if undocumented_codes and extra_codes:
        code = "undocumented-and-extra-codes"
        message = "Found undocumented and extra codes!"
    elif undocumented_codes:
        code = "undocumented-codes"
        message = "Found undocumented codes!"
    else:
        assert extra_codes
        code = "extra-codes"
        message = "Found extra codes!"

    raise DiagnosticError(
        code=code,
        message=message,
        causes=causes,
        hint_stmt=None,
    )


class _BooleanOptionalAction(argparse.Action):  # pragma: no cover
    """Handle optional boolean arguments that can be flipped with `--no-`.

    Adapted from argparse.BooleanOptionalAction on Python 3.9.0.
    """

    def __init__(
        self,
        option_strings: Sequence[str],
        dest: str,
        nargs: int | str | None = None,
        const: T | None = None,
        default: T | str | None = None,
        type: Callable[[str], T] | argparse.FileType | None = None,
        choices: Iterable[T] | None = None,
        required: bool = False,
        help: str | None = None,
        metavar: str | tuple[str, ...] | None = None,
    ):
        _option_strings: list[str] = []
        for option_string in option_strings:
            _option_strings.append(option_string)

            if option_string.startswith("--"):
                option_string = "--no-" + option_string[2:]
                _option_strings.append(option_string)

        super().__init__(
            option_strings=_option_strings,
            dest=dest,
            nargs=0,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar,
        )

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        assert option_string
        if option_string in self.option_strings:
            setattr(namespace, self.dest, not option_string.startswith("--no-"))

    def format_usage(self):
        return " | ".join(self.option_strings)


def _get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="diagnostic.check-docs",
        description=(
            "Check that all the errors in the source code have a documentation entry."
        ),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "source",
        metavar="source",
        help="Path to the source code file or directory.",
    )
    parser.add_argument(
        "docs_index",
        metavar="error-index",
        help="Path to the documentation file or directory serving as the error index.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Print detailed information about the codes discovered.",
    )
    parser.add_argument(
        "--fail-on-extra",
        dest="fail_on_extra",
        default=True,
        action=_BooleanOptionalAction,
        help=(
            "Fail if there are codes in the documentation headings, that are "
            "not in the source code."
        ),
    )
    parser.set_defaults(fail_on_extra=True)
    return parser


def main() -> None:
    """Main entry point for the script."""
    parser = _get_parser()
    args = parser.parse_args()

    source = Path(args.source)
    docs_index = Path(args.docs_index)

    if not source.exists():
        print(f"Source {source} does not exist.")
        sys.exit(1)
    if source.is_file() and not source.name.endswith(".py"):
        print(f"Source {source} is not a Python file.")
        sys.exit(1)

    if not docs_index.exists():
        print(f"Error index {docs_index} does not exist.")
        sys.exit(1)
    if docs_index.is_file() and not docs_index.name.endswith((".rst", ".md")):
        print(
            f"Error index {docs_index} is not a " "reStructuredText or Markdown file."
        )
        sys.exit(1)

    try:
        _process(source, docs_index, args.verbose, args.fail_on_extra)
    except DiagnosticError as e:
        rich.print(e)
        sys.exit(1)



"""Supporting functions for parsing the source code and documentation.
"""

import ast
import os
import sys
from collections import defaultdict
from pathlib import Path
from typing import Callable

if sys.version_info >= (3, 10):  # pragma: no cover
    from typing import TypeAlias
else:  # pragma: no cover
    from typing_extensions import TypeAlias

import docutils
import docutils.core
import docutils.nodes
import rich
from markdown_it import MarkdownIt
from rich.markup import escape

from ._base import RE_code

codeLocationMapping: TypeAlias = "dict[str, list[tuple[Path, int]]]"


def _ignoring(*, ctx: str, what: str, why: str, where: "tuple[Path, int]") -> None:
    rich.print(f"[yellow]Ignoring {escape(what)}[/]")
    rich.print(f"  [magenta]{escape(str(where[0]))}[/]:[cyan]{where[1]}[/]")
    rich.print(f"  [blue]{ctx}[/]: {escape(why)}")


def handle_directory_traversal(
    path: Path,
    func: Callable[[Path], codeLocationMapping],
    *,
    extensions: "tuple[str, ...]",
) -> codeLocationMapping:
    if not path.is_dir():
        assert path.name.endswith(
            extensions
        ), f"expected {path} to end with one of {extensions}"
        return func(path)

    codes: codeLocationMapping = defaultdict(list)
    for dirpath, _, filenames in os.walk(path):
        for filename in filenames:
            if filename.endswith(extensions):
                these_codes = func(Path(dirpath) / filename)
                for code, locations in these_codes.items():
                    codes[code].extend(locations)
    return codes


def find_codes_in_sources(src_path: Path) -> codeLocationMapping:
    """Find all the codes in the source code, using the AST.

    This uses the AST to find all the error codes in the source code. An
    error code is found in two ways:

    - A class with a `code` attribute, which is a string literal.
    - A call with a `code` keyword argument, which is a string literal.

    Returns:
        A dictionary mapping the code to a list of (filename, line number)
        tuples, where the code was found.
    """

    def find_codes_in_file(file: Path) -> codeLocationMapping:
        codes: defaultdict[str, list[tuple[Path, int]]] = defaultdict(list)

        with open(file) as f:
            tree = ast.parse(f.read())

        print(ast.dump(tree))
        for node in ast.walk(tree):
            if isinstance(node, ast.ClassDef):
                for attr in node.body:
                    if (
                        isinstance(attr, ast.Assign)
                        and len(attr.targets) == 1
                        and isinstance(attr.targets[0], ast.Name)
                        and attr.targets[0].id == "code"
                        and isinstance(attr.value, ast.Str)
                    ):
                        ref = attr.value.s
                        if not RE_code.match(ref):
                            _ignoring(
                                ctx="class-attribute",
                                what=f"{ref!r}",
                                why="not a valid code",
                                where=(file, node.lineno),
                            )
                            continue
                        codes[ref].append((file, node.lineno))
            elif isinstance(node, ast.Call):
                print(ast.dump(node))
                for kw in node.keywords:
                    if kw.arg == "code" and isinstance(kw.value, ast.Str):
                        ref = kw.value.s
                        if not RE_code.match(ref):
                            _ignoring(
                                ctx="call-argument",
                                what=f"{ref!r}",
                                why="not a valid code",
                                where=(file, node.lineno),
                            )
                            continue
                        codes[ref].append((file, node.lineno))

        return codes

    return handle_directory_traversal(src_path, find_codes_in_file, extensions=(".py",))


def find_code_headings_in_document(doc_path: Path) -> codeLocationMapping:
    """Finds all the level 2+ headings within the document.

    Returns:
        A dictionary mapping the code to a list of line numbers, where the
        heading was found.
    """

    def find_codes_in_file(doc_path: Path) -> codeLocationMapping:
        if doc_path.name.endswith(".md"):
            return find_code_headings_in_markdown(doc_path)
        else:
            assert doc_path.name.endswith(".rst")
            return find_code_headings_in_rst(doc_path)

    return handle_directory_traversal(
        doc_path, find_codes_in_file, extensions=(".md", ".rst")
    )


def find_code_headings_in_markdown(doc_path: Path) -> codeLocationMapping:
    """Finds potential code headings in a Markdown document.

    Returns:
        A dictionary mapping the code to a list of line numbers, where the
        heading was found.
    """
    parser = MarkdownIt()
    tokens = parser.parse(Path(doc_path).read_text())  # type: ignore

    found_headings: list[tuple[str, int]] = []
    current_heading: "tuple[str, int] | None" = None
    for token in tokens:
        if token.type == "heading_open":
            assert not current_heading, "active heading already"
            current_heading = ("", -1)
        elif token.type == "heading_close":
            assert current_heading, "no active heading"
            # Skip headings that are not codes
            if RE_code.match(current_heading[0]):
                found_headings.append(current_heading)
            current_heading = None
        elif current_heading is not None:
            assert token.type == "inline", "expected inline token"
            assert token.map, "expected line info"
            new_text = current_heading[0] + token.content
            assert isinstance(new_text, str)
            current_heading = (new_text, token.map[0] + 1)

    codes: codeLocationMapping = defaultdict(list)
    for heading, lineno in found_headings:
        codes[heading].append((doc_path, lineno))  # pyright: ignore

    return codes


def find_code_headings_in_rst(doc_path: Path) -> codeLocationMapping:
    """Finds all the level 2+ headings within the document.

    Returns:
        A dictionary mapping the code to a list of line numbers, where the
        heading was found.
    """

    with open(doc_path) as f:
        rst_string = f.read()

    # Parse the reStructuredText document
    document: docutils.nodes.document = docutils.core.publish_doctree(rst_string)

    # Iterate through the document and extract all headings
    codes: codeLocationMapping = defaultdict(list)
    for node in document.findall(
        condition=lambda n: getattr(n, "tagname", None) == "title"
    ):
        heading_text = node.astext()

        # Determine the level of the heading
        heading_level = -1  # we'll always have `document` as the parent
        tree = node
        while tree.parent is not None:
            tree = tree.parent
            heading_level += 1

        # Skip headings that are not codes
        if not RE_code.match(heading_text):
            continue

        assert node.line
        codes[heading_text].append((doc_path, node.line))

    return codes
